@@ -54,7 +54,7 @@ struct client_window {
 	int x, y, w, h;
 
 	/* -1 = sticky */
-	int desk;
+	int page;
 
 	enum window_type type;
 	char *name;
@@ -72,21 +72,26 @@ struct pager {
 	Pixmap pixmap;
 	GC active_win_gc;
 	GC inactive_win_gc;
-	GC active_desk_gc;
-	GC inactive_desk_gc;
+	GC active_page_gc;
+	GC inactive_page_gc;
 	GC win_border_gc;
+	GC pager_border_gc;
 	GC grid_gc;
+	GC desk_grid_gc;
 
 	struct client_window *windows;
 	int nr_windows;
 
 	Window active_win;
-	int active_desk;
+	int active_page;
+        int border_width;
 
-	int cols, rows;
+	/* cols, rows, desk_cols, desk_rows, page_cols, page_rows */
+	int size[6];
 
 	int x, y;
 	int w, h;
+	double zoom;
 
 	/* user set geometry, these are constants
 	 *
@@ -101,8 +106,8 @@ struct pager {
 
 	int root_w;
 	int root_h;
-	int desk_w;
-	int desk_h;
+	int page_w;
+	int page_h;
 
 	int w_extra;
 	int h_extra;
@@ -154,6 +159,8 @@ struct pager {
 	XftFont *popup_font;
 };
 
+static int netwm_compatibility = 0;
+
 static GC make_gc(Window window, const char *color)
 {
 	GC gc;
@@ -231,12 +238,13 @@ static void pager_update_strut(struct pager *pager)
 
 static void pager_configure(struct pager *pager)
 {
-	int x, y;
+	int x, y, w, h;
 
 	if (x_window_get_geometry(pager->window, &x, &y, &pager->w, &pager->h)) {
 		d_print("x_window_get_geometry failed\n");
 		return;
 	}
+
 	pager_calc_x_y(pager);
 	if (x != pager->x || y != pager->y) {
 		d_print("forcing position\n");
@@ -246,15 +254,18 @@ static void pager_configure(struct pager *pager)
 	pager->needs_configure = 0;
 	pager->needs_update = 1;
 
-	pager->desk_w = (pager->w - (pager->cols - 1)) / pager->cols;
-	pager->desk_h = (pager->h - (pager->rows - 1)) / pager->rows;
-	pager->w_extra = pager->w - pager->cols * pager->desk_w - (pager->cols - 1);
-	pager->h_extra = pager->h - pager->rows * pager->desk_h - (pager->rows - 1);
+        w = pager->w - 2 * pager->border_width - pager->size[COLS] + 1;
+        h = pager->h - 2 * pager->border_width - pager->size[ROWS] + 1;
+
+	pager->page_w = w / pager->size[COLS];
+	pager->page_h = h / pager->size[ROWS];
+
+	pager->w_extra = w - pager->size[COLS] * pager->page_w;
+	pager->h_extra = h - pager->size[ROWS] * pager->page_h;
 	XFreePixmap(display, pager->pixmap);
 	pager->pixmap = XCreatePixmap(display,
 			pager->window,
-			pager->w,
-			pager->h,
+			pager->w, pager->h,
 			DefaultDepth(display, DefaultScreen(display)));
 	XSetWindowBackgroundPixmap(display, pager->window, pager->pixmap);
 	pager_update_strut(pager);
@@ -262,31 +273,36 @@ static void pager_configure(struct pager *pager)
 
 static void pager_calc_w_h(struct pager *pager)
 {
-	pager->w = pager->gw;
-	pager->h = pager->gh;
+        pager->w = (int) (pager->gw * pager->zoom + 0.5);
 	if (!pager->gh)
-		pager->h = pager->w * (pager->root_h * pager->rows) / (pager->root_w * pager->cols);
+		pager->h = pager->w * (pager->root_h * pager->size[ROWS]) / (pager->root_w * pager->size[COLS]);
+	else
+		pager->h = (int) (pager->gh * pager->zoom + 0.5);
+
 }
 
 static void pager_update_aspect(struct pager *pager)
 {
 	if (!pager->gh)
 		x_window_set_aspect(pager->window,
-				pager->root_w * pager->cols,
-				pager->root_h * pager->rows);
+				pager->root_w * pager->size[COLS],
+				pager->root_h * pager->size[ROWS]);
 }
 
 static void update_desktop_count(struct pager *pager)
 {
-	int rows = pager->rows;
-	int cols = pager->cols;
+	int rows = pager->size[ROWS];
+	int cols = pager->size[COLS];
+	int width = -1;
+	int height = -1;
 
-	if (x_get_desktop_layout(&cols, &rows)) {
+	if (netwm_compatibility &&
+	    x_get_desktop_layout(&cols, &rows, &width, &height)) {
 		d_print("x_get_desktop_layout failed\n");
 	}
-	if (rows != pager->rows || cols != pager->cols) {
-		pager->rows = rows;
-		pager->cols = cols;
+	if (rows != pager->size[ROWS] || cols != pager->size[COLS]) {
+		pager->size[ROWS] = rows;
+		pager->size[COLS] = cols;
 		pager_update_aspect(pager);
 
 		pager_calc_w_h(pager);
@@ -299,8 +315,7 @@ static void update_desktop_count(struct pager *pager)
 		XFreePixmap(display, pager->pixmap);
 		pager->pixmap = XCreatePixmap(display,
 				pager->window,
-				pager->w,
-				pager->h,
+				pager->w, pager->h,
 				DefaultDepth(display, DefaultScreen(display)));
 
 		pager->needs_configure = 1;
@@ -327,7 +342,7 @@ static void pager_update_properties(struct pager *pager)
 	Window move_win = -1, popup_win = -1;
 	Window *windows;
 	int nr_windows;
-	int i, j;
+	int i, j, active_col, active_row;
 
 	update_desktop_count(pager);
 
@@ -340,6 +355,12 @@ static void pager_update_properties(struct pager *pager)
 	pager->needs_update_properties = 0;
 	pager->needs_update = 1;
 
+	active_col = pager->active_page % pager->size[COLS];
+	active_row = pager->active_page / pager->size[COLS];
+	active_col = active_col % pager->size[PAGE_COLS];
+	active_row = active_row % pager->size[PAGE_ROWS];
+	d_print("active page=(%d,%d)\n", active_col, active_row);
+
 	if (pager->mouse.window_idx != -1)
 		move_win = pager->windows[pager->mouse.window_idx].window;
 	if (pager->popup_idx != -1)
@@ -348,6 +369,8 @@ static void pager_update_properties(struct pager *pager)
 	pager_free_windows(pager);
 
 	pager->nr_windows = nr_windows;
+	d_print("nr_windows: %d\n", nr_windows);
+
 	pager->windows = xnew(struct client_window, pager->nr_windows);
 
 	j = 0;
@@ -355,7 +378,6 @@ static void pager_update_properties(struct pager *pager)
 		struct client_window *win = &pager->windows[j];
 
 		win->window = windows[i];
-
 		ignore_bad_window = 1;
 
 		win->type = WINDOW_TYPE_NORMAL;
@@ -366,8 +388,8 @@ static void pager_update_properties(struct pager *pager)
 
 		win->states = 0;
 		if (x_window_get_states(win->window, &win->states)) {
-/* 			fprintf(stderr, "could not get states of window 0x%x\n", (int)windows[i]); */
-			continue;
+ 			fprintf(stderr, "could not get states of window 0x%x, assuming 'MODAL'\n", (int)windows[i]); 
+			win->states = WINDOW_STATE_MODAL;
 		}
 
 		ignore_bad_window = 0;
@@ -376,16 +398,13 @@ static void pager_update_properties(struct pager *pager)
 /* 			d_print("skip pager 0x%x\n", (int)windows[i]); */
 			continue;
 		}
-		win->desk = -1;
-		if (x_window_get_desktop(win->window, &win->desk)) {
+		win->page = -1;
+
+		if (x_window_get_desktop(win->window, active_col, active_row, &win->page, &win->x, &win->y, &win->w, &win->h, pager->size)) {
 			fprintf(stderr, "could not get desktop of window 0x%x\n", (int)win->window);
 			continue;
 		}
 
-		if (x_window_get_geometry(win->window, &win->x, &win->y, &win->w, &win->h)) {
-			fprintf(stderr, "could not get geometry of window 0x%x\n", (int)win->window);
-			continue;
-		}
 		if (x_window_get_title(win->window, &win->name)) {
 			fprintf(stderr, "could not get name of window 0x%x\n", (int)win->window);
 			/* continue; */
@@ -394,8 +413,9 @@ static void pager_update_properties(struct pager *pager)
 		win->icon_w = -1;
 		win->icon_h = -1;
 		win->icon_data = NULL;
-/* 		d_print("new window 0x%x '%s'\n", (int)win->window, win->name); */
-
+		/*
+ 		d_print("new window 0x%x '%s' page %d\n", (int)win->window, win->name, win->page); 
+		*/
 		/* FIXME: breaks sometimes */
 /* 		XSelectInput(display, win->window, PropertyChangeMask); */
 		j++;
@@ -409,13 +429,18 @@ static void pager_update_properties(struct pager *pager)
 	if (popup_win != -1)
 		pager->popup_idx = get_window_index(pager, popup_win);
 
-	x_get_current_desktop(&pager->active_desk);
+	x_get_current_desktop(&pager->active_page, pager->size);
 	x_get_active_window(&pager->active_win);
 }
 
 static void do_draw_window(struct pager *pager, struct client_window *window, int px, int py, int pw, int ph)
 {
 	GC gc;
+	XftColor *color;
+	XGlyphInfo extents;
+	XRectangle ra;
+	int len = strlen(window->name);
+	int x, y, w, h;
 
 	if (pager->active_win == window->window) {
 		gc = pager->active_win_gc;
@@ -424,51 +449,45 @@ static void do_draw_window(struct pager *pager, struct client_window *window, in
 	}
 
 	/* NOTE: XDrawRectangle draws rectangle one pixel larger than XFillRectangle */
-	XDrawRectangle(display, pager->pixmap, pager->win_border_gc, px, py, pw - 1, ph - 1);
-
-	px++;
-	py++;
-	pw -= 2;
-	ph -= 2;
+   
 	XFillRectangle(display, pager->pixmap, gc, px, py, pw, ph);
+	XDrawRectangle(display, pager->pixmap, pager->win_border_gc, px, py, pw, ph);
 
-	if (pager->show_window_titles) {
-		XftColor *color;
-		XGlyphInfo extents;
-		XRectangle ra;
-		int len = strlen(window->name);
-		int x, y, w, h;
+	if (! pager->show_window_titles)
+		return;
 
-		ra.x = px + TITLE_HPAD;
-		ra.y = py + TITLE_VPAD;
-		w = pw - 2 * TITLE_HPAD;
-		h = ph - 2 * TITLE_VPAD;
+	if (!opt_transient && pager->zoom * pager->zoom < zoom)
+		return;
 
-		if (w < 1 || h < 1)
-			return;
+	ra.x = px + TITLE_HPAD;
+	ra.y = py + TITLE_VPAD;
+	w = pw - 2 * TITLE_HPAD;
+	h = ph - 2 * TITLE_VPAD;
 
-		ra.width = w;
-		ra.height = h;
+	if (w < 1 || h < 1)
+		return;
 
-		XftTextExtentsUtf8(display, pager->window_font, (FcChar8 *)window->name, len, &extents);
+	ra.width = w;
+	ra.height = h;
+
+	XftTextExtentsUtf8(display, pager->window_font, (FcChar8 *)window->name, len, &extents);
 /* 		d_print("w = %d, h = %d, x = %d, y = %d, xoff = %d, yoff = %d\n", extents.width, extents.height, extents.x, extents.y, extents.xOff, extents.yOff); */
 
-		x = ra.x + extents.x;
-		y = ra.y + extents.y;
-		if (extents.width < ra.width)
-			x += (ra.width - extents.width) / 2;
-		y += (ra.height - extents.height) / 2;
+	x = ra.x + extents.x;
+	y = ra.y + extents.y;
+	if (extents.width < ra.width)
+		x += (ra.width - extents.width) / 2;
+	y += (ra.height - extents.height) / 2;
 
-		if (pager->active_win == window->window) {
-			color = &pager->active_win_font_color;
-		} else {
-			color = &pager->inactive_win_font_color;
-		}
-		XftDrawSetClipRectangles(pager->xft_draw, 0, 0, &ra, 1);
-		XftDrawStringUtf8(pager->xft_draw, color,
-				pager->window_font, x, y,
-				(FcChar8 *)window->name, len);
+	if (pager->active_win == window->window) {
+		color = &pager->active_win_font_color;
+	} else {
+		color = &pager->inactive_win_font_color;
 	}
+	XftDrawSetClipRectangles(pager->xft_draw, 0, 0, &ra, 1);
+	XftDrawStringUtf8(pager->xft_draw, color,
+			  pager->window_font, x, y,
+			  (FcChar8 *)window->name, len);
 }
 
 static void draw_window(struct pager *pager, struct client_window *window)
@@ -478,11 +497,14 @@ static void draw_window(struct pager *pager, struct client_window *window)
 	double x_scale, y_scale;
 	int px, py, pw, ph;
 
-	x_scale = (double)pager->root_w / (double)pager->desk_w;
-	y_scale = (double)pager->root_h / (double)pager->desk_h;
+	d_print(" | %d: %s %d %d", window->page, window->name, window->x, window->y);
 
-	px = (int)((double)window->x / x_scale);
-	py = (int)((double)window->y / y_scale);
+	x_scale = ((double)pager->root_w - win_shift_x) / ((double)(pager->page_w + 1));
+	y_scale = ((double)pager->root_h - win_shift_y) / ((double)(pager->page_h + 1));
+
+	px = (int)((double)(window->x - win_shift_x) / x_scale) + pager->border_width;
+	py = (int)((double)(window->y - win_shift_y) / y_scale) + pager->border_width;
+        
 	pw = (int)((double)window->w / x_scale);
 	if (window->states & WINDOW_STATE_SHADED) {
 		ph = (int)((double)WINDOW_SHADED_H / y_scale);
@@ -497,25 +519,26 @@ static void draw_window(struct pager *pager, struct client_window *window)
 
 	XftDrawChange(pager->xft_draw, pager->pixmap);
 
-	if (window->desk == -1) {
+	if (window->page == -1) {
 		if (!pager->show_sticky)
 			return;
 
-		for (row = 0; row < pager->rows; row++) {
-			for (col = 0; col < pager->cols; col++) {
-				x_off = col * pager->desk_w + col;
-				y_off = row * pager->desk_h + row;
+		d_print(" | %d: %d %d", window->page, window->x, window->y);
+		for (row = 0; row < pager->size[ROWS]; row++) {
+			for (col = 0; col < pager->size[COLS]; col++) {
+				x_off = col * (pager->page_w + 1);
+				y_off = row * (pager->page_h + 1);
 
 				do_draw_window(pager, window, px + x_off, py + y_off, pw, ph);
 			}
 		}
 	} else {
-		col = window->desk % pager->cols;
-		row = window->desk / pager->cols;
-
-		x_off = col * pager->desk_w + col;
-		y_off = row * pager->desk_h + row;
-
+		col = window->page % pager->size[COLS];
+		row = window->page / pager->size[COLS];
+		col = (col / pager->size[PAGE_COLS]) * pager->size[PAGE_COLS];
+		row = (row / pager->size[PAGE_ROWS]) * pager->size[PAGE_ROWS];
+		x_off = col * (pager->page_w + 1) + col;
+		y_off = row * (pager->page_h + 1) + row;
 		do_draw_window(pager, window, px + x_off, py + y_off, pw, ph);
 	}
 }
@@ -523,31 +546,42 @@ static void draw_window(struct pager *pager, struct client_window *window)
 /* convert mouse coordinates (@x, @y) (which are relative to the pager window)
  * to root window coordinates (@rx, @ry) and desktop value (@desk)
  */
-static void pager_coords_to_root(struct pager *pager, int x, int y, int *rx, int *ry, int *desk)
+static void pager_coords_to_root(struct pager *pager, int x, int y, int *rx, int *ry, int *page)
 {
-	int row, col;
+	int col, row, w, h;
 
-	col = x / (pager->desk_w + 1);
-	row = y / (pager->desk_h + 1);
+	x -= pager->border_width;
+        if (x < 0) x = 0;
+        y -= pager->border_width;
+        if (y < 0) y = 0;
 
-	x -= col * (pager->desk_w + 1);
-	y -= row * (pager->desk_h + 1);
+	w = pager->page_w + 1;
+	h = pager->page_h + 1;
 
-	*rx = x * pager->root_w / pager->desk_w;
-	*ry = y * pager->root_h / pager->desk_h;
-	*desk = row * pager->cols + col;
+	col = x / w;
+	x -= col * w;
+	row = y / h;
+	y -= row * h;
+
+	*page =  row * pager->size[COLS] + col;
+
+	col = col % pager->size[PAGE_COLS];
+	row = row % pager->size[PAGE_ROWS];
+
+	*rx = (x * pager->root_w) / pager->page_w + col * pager->root_w;
+	*ry = (y * pager->root_h) / pager->page_h + row * pager->root_h;
 }
 
-static int get_window_idx(struct pager *pager, int rx, int ry, int desk)
+static int get_window_idx(struct pager *pager, int rx, int ry, int page)
 {
-	double scale = (double)pager->root_w / (double)pager->desk_w;
+	double scale = (double)pager->root_w / (double)pager->page_w;
 	int i;
 
 	for (i = pager->nr_windows - 1; i >= 0; i--) {
 		struct client_window *window = &pager->windows[i];
 		int w, h, sw, sh;
 
-		if (!(window->desk == desk || (window->desk == -1 && pager->show_sticky)))
+		if (!(window->page == page || (window->page == -1 && pager->show_sticky)))
 			continue;
 
 		if (window->type == WINDOW_TYPE_DESKTOP)
@@ -583,54 +617,66 @@ static void pager_update(struct pager *pager)
 	pager->needs_update = 0;
 
 	/* desktops */
-	for (row = 0; row < pager->rows; row++) {
-		y = row * (pager->desk_h + 1);
-		for (col = 0; col < pager->cols; col++) {
-			x = col * (pager->desk_w + 1);
-			if (row * pager->cols + col == pager->active_desk) {
+	for (row = 0; row < pager->size[ROWS]; row++) {
+		y = row * (pager->page_h + 1) + pager->border_width;
+		for (col = 0; col < pager->size[COLS]; col++) {
+			x = col * (pager->page_w + 1) + pager->border_width;
+			if (row * pager->size[COLS] + col == pager->active_page) {
 				XFillRectangle(display, pager->pixmap,
-						pager->active_desk_gc,
-						x, y, pager->desk_w, pager->desk_h);
+						pager->active_page_gc,
+						x, y, pager->page_w+1, pager->page_h+1);
 			} else {
 				XFillRectangle(display, pager->pixmap,
-						pager->inactive_desk_gc,
-						x, y, pager->desk_w, pager->desk_h);
+						pager->inactive_page_gc,
+						x, y, pager->page_w+1, pager->page_h+1);
 			}
 		}
 	}
 
 	if (pager->w_extra) {
 		XFillRectangle(display, pager->pixmap,
-				pager->inactive_desk_gc,
-				pager->w - pager->w_extra,
-				0,
+				pager->inactive_page_gc,
+				pager->w - pager->w_extra + pager->border_width,
+				pager->border_width,
 				pager->w_extra,
 				pager->h);
 	}
 
 	if (pager->h_extra) {
 		XFillRectangle(display, pager->pixmap,
-				pager->inactive_desk_gc,
-				0,
-				pager->h - pager->h_extra,
+				pager->inactive_page_gc,
+				pager->border_width,
+				pager->h - pager->h_extra + pager->border_width,
 				pager->w - pager->w_extra,
 				pager->h_extra);
 	}
 
+	/* border of pager */
+        for (i = 0; i<pager->border_width; i++)
+		XDrawRectangle(display, pager->pixmap, pager->pager_border_gc, i, i, pager->w -2*i-1, pager->h - 2*i-1);
+
+
 	/* grid */
-	for (row = 1; row < pager->rows; row++) {
-		y = row * (pager->desk_h + 1) - 1;
-		XDrawLine(display, pager->pixmap, pager->grid_gc, 0, y, pager->w, y);
+	x = pager->size[ROWS] / pager->size[DESK_ROWS];
+	for (row = 1; row < pager->size[ROWS]; row++) {
+		y = row * (pager->page_h + 1) - 1;
+		i = row % x;
+		XDrawLine(display, pager->pixmap, (i)? pager->grid_gc:pager->desk_grid_gc, pager->border_width, y+ pager->border_width, pager->w - pager->border_width, y + pager->border_width);
 	}
-	for (col = 1; col < pager->cols; col++) {
-		x = col * (pager->desk_w + 1) - 1;
-		XDrawLine(display, pager->pixmap, pager->grid_gc, x, 0, x, pager->h);
+
+	y = pager->size[COLS] / pager->size[DESK_COLS];
+	for (col = 1; col < pager->size[COLS]; col++) {
+		x = col * (pager->page_w + 1) - 1;
+		i = col % y;
+		XDrawLine(display, pager->pixmap, (i)? pager->grid_gc:pager->desk_grid_gc, x + pager->border_width, pager->border_width, x + pager->border_width, pager->h - pager->border_width);
 	}
 
 	if (x_get_showing_desktop(&showing_desktop)) {
 	}
 	if (!showing_desktop) {
 		/* windows */
+		if (pager->nr_windows)
+			d_print("windows:");
 		for (i = 0; i < pager->nr_windows; i++) {
 			switch (pager->windows[i].type) {
 			case WINDOW_TYPE_DESKTOP:
@@ -647,9 +693,11 @@ static void pager_update(struct pager *pager)
 				break;
 			}
 		}
+		d_print("\n");
 	}
 
 	XClearWindow(display, pager->window);
+
 /* 	XFlush(display); */
 }
 
@@ -685,9 +733,9 @@ static int cursor_to_desk(struct pager *pager, int x, int y)
 {
 	int row, col;
 
-	col = (x - pager->x) / (pager->desk_w + 1);
-	row = (y - pager->y) / (pager->desk_h + 1);
-	return row * pager->cols + col;
+	col = (x - pager->x) / (pager->page_w + 1);
+	row = (y - pager->y) / (pager->page_h + 1);
+	return row * pager->size[COLS] + col;
 }
 
 /* show popup window below (or above) the window in the pager */
@@ -702,9 +750,11 @@ static void popup_show(struct pager *pager, int cx, int cy)
 
 	if (!pager->show_popups)
 		return;
+	if (!opt_transient && pager->zoom * pager->zoom < zoom)
+		return;
 	
 	win = &pager->windows[pager->popup_idx];
-	win_row = cursor_to_desk(pager, cx, cy) / pager->cols;
+	win_row = cursor_to_desk(pager, cx, cy) / pager->size[COLS];
 	len = strlen(win->name);
 
 	bw = 1;
@@ -724,11 +774,11 @@ static void popup_show(struct pager *pager, int cx, int cy)
 		x = x_max;
 
 	/* y = 4px below (or above) the window in the pager */
-	y = pager->y + win_row * pager->desk_h + win_row +
-		(win->y + win->h) * pager->desk_h / pager->root_h + 4;
+	y = pager->y + win_row * pager->page_h + win_row +
+		(win->y + win->h) * pager->page_h / pager->root_h + 4;
 	if (y + h > y_max)
-		y = pager->y + win_row * pager->desk_h + win_row +
-			win->y * pager->desk_h / pager->root_h - h - 4;
+		y = pager->y + win_row * pager->page_h + win_row +
+			win->y * pager->page_h / pager->root_h - h - 4;
 	if (y < y_min)
 		y = y_min;
 	XMoveResizeWindow(display, pager->popup_window, x, y, w, h);
@@ -753,10 +803,12 @@ static void popup_hide(struct pager *pager)
 
 char *active_win_color = "rgb:99/a6/bf";
 char *inactive_win_color = "rgb:f6/f6/f6";
-char *active_desk_color = "rgb:65/72/8c";
-char *inactive_desk_color = "rgb:ac/ac/ac";
+char *active_page_color = "rgb:65/72/8c";
+char *inactive_page_color = "rgb:ac/ac/ac";
 char *win_border_color = "rgb:00/00/00";
+char *pager_border_color = "rgb:5f/5f/5f";
 char *grid_color = "rgb:ff/ff/ff";
+char *desk_grid_color = "rgb:00/ff/ff";
 
 char *active_win_font_color = "rgb:00/00/00";
 char *inactive_win_font_color = "rgb:00/00/00";
@@ -764,19 +816,115 @@ char *inactive_win_font_color = "rgb:00/00/00";
 char *popup_color = "rgb:e6/e6/e6";
 char *popup_font_color = "rgb:00/00/00";
 
-struct pager *pager_new(const char *geometry, int cols, int rows)
+int border_width = 0;
+int win_shift_x = 0;
+int win_shift_y = 0;
+int win_drift_x = 0;
+int win_drift_y = 0;
+int switch_delay = 0;
+double zoom = 1.0;
+double zoom_factor = 1.04663513939;  /* 1.2^0.25 */
+
+static int set_pager_grid(struct pager *pager,
+			   int cols, int rows, int desk_cols, int desk_rows)
+{
+	int wm_c = 1, wm_r = 1, width = 0, height = 0;
+	int x, y, failed = 0;
+
+        netwm_compatibility = x_is_netwm_compatible_wm_running();
+	d_print("NetWM compatibility: %d%s\n", 
+		netwm_compatibility,
+		netwm_compatibility? "": " (?)");
+
+	/* NetWM compatible window manager must be running */
+	while (1) {
+		static int count = 0;
+		if (x_get_desktop_layout(&wm_c, &wm_r, &width, &height) == 0) {
+			d_print("Got desktop layout (%dx%d) width=%d height=%d [%dx%d]\n"
+			"NetWM compatible wm should be running\n",
+			wm_c, wm_r, width, height,
+			width/DisplayWidth(display, DefaultScreen(display)),
+			height/DisplayHeight(display, DefaultScreen(display)));
+			break;
+		}
+		if (count == 0)
+			fprintf(stderr, "NetWM compatible window manager not running. Waiting at most %d seconds.\n", WM_WAIT);
+		if (count == WM_WAIT) {
+			failed = 1;
+			fprintf(stderr, "Couldn't get desktop layout.\n");
+			fprintf(stderr, "Exiting.\n");
+			break;
+		}
+		sleep(1);
+		count++;
+	}
+
+        if (cols == -1)
+		cols = width / DisplayWidth(display, DefaultScreen(display));
+        if (rows == -1)
+		rows = height / DisplayHeight(display, DefaultScreen(display));
+        if (cols < 1)
+		cols = 1;
+        if (rows < 1)
+		rows = 1;
+        if (cols > 32)
+		cols = 32;
+        if (rows > 32)
+		rows = 32;
+
+	if (desk_cols == -1)
+		desk_cols = wm_c;
+	if (desk_rows == -1)
+		desk_rows = wm_r;
+	if (desk_cols < 1)
+		desk_cols = 1;
+	if (desk_rows < 1)
+		desk_rows = 1;
+	if (desk_cols > 32)
+		desk_cols = 32;
+	if (desk_rows > 32)
+		desk_rows = 32;
+
+	if (failed) {
+		/* set desktop layout */
+		width = cols * DisplayWidth(display, DefaultScreen(display));
+		height = cols * DisplayHeight(display, DefaultScreen(display));
+		if (x_set_desktop_layout(desk_cols, desk_rows, width, height)) {
+		  fprintf(stderr, "Couldn't set desktop layout (%d x %d) width=%d height=%d.\n", desk_cols, desk_rows, width, height);
+			return 1;
+		}
+	}
+
+	if (x_window_get_geometry(DefaultRootWindow(display), &x, &y, &pager->root_w, &pager->root_h)) {
+		d_print("x_window_get_geometry for root window failed\n");
+		free(pager);
+		return 1;
+	}
+
+	pager->size[PAGE_COLS] = cols;
+	pager->size[PAGE_ROWS] = rows;
+	pager->size[DESK_COLS] = desk_cols;
+	pager->size[DESK_ROWS] = desk_rows;
+	pager->size[COLS] = cols * desk_cols;
+	pager->size[ROWS] = rows * desk_rows;
+
+	return 0;
+}
+
+struct pager *pager_new(const char *geometry,
+			int cols, int rows, int desk_cols, int desk_rows)
 {
 	unsigned long popup_bg;
 
 	struct pager *pager;
 	XSetWindowAttributes attrib;
 	unsigned long attrib_mask;
-	int x, y;
 	Visual *visual;
 	Colormap cm;
 	int gflags, gx, gy;
 	unsigned int gw, gh;
-	int wm_c = -1, wm_r = -1;
+
+	pager = xnew(struct pager, 1);
 
 	x_parse_geometry(geometry, &gflags, &gx, &gy, &gw, &gh);
 	if (!(gflags & XValue))
@@ -792,74 +940,32 @@ struct pager *pager_new(const char *geometry, int cols, int rows)
 	if (gy < 0)
 		gy *= -1;
 
-	/* NetWM compatible window manager must be running */
-	while (1) {
-		static int count = 0;
-
-		if (x_is_netwm_compatible_wm_running())
-			break;
-		if (x_get_desktop_layout(&wm_c, &wm_r) == 0) {
-			d_print("Got desktop layout (%dx%d), NetWM compatible wm should be running\n", wm_c, wm_r);
-			break;
-		}
-		if (count == 0)
-			fprintf(stderr, "NetWM compatible window manager not running. Waiting at most %d seconds.\n", WM_WAIT);
-		if (count == WM_WAIT) {
-			fprintf(stderr, "Exiting.\n");
-			return NULL;
-		}
-		sleep(1);
-		count++;
-	}
-
-	if (cols == -1 || rows == -1) {
-		/* use desktop layout set by WM */
-		if (wm_c == -1 || wm_r == -1) {
-			if (x_get_desktop_layout(&wm_c, &wm_r)) {
-				fprintf(stderr, "Couldn't get desktop layout.\n");
-				return NULL;
-			}
-		}
-		cols = wm_c;
-		rows = wm_r;
-	} else {
-		/* set desktop layout */
-		if (cols > 32)
-			cols = 32;
-		if (rows > 32)
-			rows = 32;
-		if (x_set_desktop_layout(cols, rows)) {
-			fprintf(stderr, "Couldn't set desktop layout (%dx%d).\n", cols, rows);
-			return NULL;
-		}
-	}
-
-	pager = xnew(struct pager, 1);
-
-	if (x_window_get_geometry(DefaultRootWindow(display), &x, &y, &pager->root_w, &pager->root_h)) {
-		d_print("x_window_get_geometry for root window failed\n");
+	if (set_pager_grid(pager, cols, rows, desk_cols, desk_rows)) {
 		free(pager);
 		return NULL;
 	}
 
-	pager->cols = cols;
-	pager->rows = rows;
-
-	d_print("g: %d,%d %d, %d %d\n", gx, gy, gw, gflags & XNegative, gflags & YNegative);
-
 	pager->gx = gx;
 	pager->gy = gy;
 	pager->gx_negative = (gflags & XNegative) != 0;
 	pager->gy_negative = (gflags & YNegative) != 0;
+	pager->zoom = (opt_zoom)? zoom : 1.0;
 	pager->gw = gw;
 	pager->gh = gh;
 	pager_calc_w_h(pager);
 	pager_calc_x_y(pager);
 
+	d_print("geometry: %s -> %dx%d+%d+%d flags: %d %d pages: %dx%d desks: %dx%d\n", 
+		geometry, pager->w, pager->h, pager->x, pager->y,
+		gflags & XNegative, gflags & YNegative,
+		pager->size[PAGE_COLS], pager->size[PAGE_ROWS],
+		pager->size[DESK_COLS], pager->size[DESK_ROWS]);
+
 	pager->windows = NULL;
 	pager->nr_windows = 0;
 	pager->active_win = 0;
-	pager->active_desk = 0;
+	pager->active_page = 0;
+	pager->border_width = border_width;
 
 	pager->popup_idx = -1;
 	pager->popup_visible = 0;
@@ -876,7 +982,7 @@ struct pager *pager_new(const char *geometry, int cols, int rows)
 	pager->allow_cover = 1;
 
 	pager->opacity = 1.0;
-	pager->layer = LAYER_NORMAL;
+	pager->layer = LAYER_COMMUTE;
 
 	pager->mouse.window_idx = -1;
 	pager->mouse.button = -1;
@@ -892,7 +998,8 @@ struct pager *pager_new(const char *geometry, int cols, int rows)
 		PointerMotionMask | EnterWindowMask | LeaveWindowMask |
 		ExposureMask;
 	pager->window = XCreateWindow(display, DefaultRootWindow(display),
-			pager->x, pager->y, pager->w, pager->h,
+			pager->x, pager->y,
+                        pager->w, pager->h,
 			0, // border
 			CopyFromParent,
 			InputOutput,
@@ -906,7 +1013,7 @@ struct pager *pager_new(const char *geometry, int cols, int rows)
 	pager->popup_window = XCreateWindow(display, DefaultRootWindow(display),
 			0, 0,
 			1, 1,
-			1, // border
+			1, /* border */
 			CopyFromParent,
 			InputOutput,
 			CopyFromParent,
@@ -928,10 +1035,12 @@ struct pager *pager_new(const char *geometry, int cols, int rows)
 
 	pager->active_win_gc    = make_gc(pager->window, active_win_color);
 	pager->inactive_win_gc  = make_gc(pager->window, inactive_win_color);
-	pager->active_desk_gc   = make_gc(pager->window, active_desk_color);
-	pager->inactive_desk_gc = make_gc(pager->window, inactive_desk_color);
+	pager->active_page_gc   = make_gc(pager->window, active_page_color);
+	pager->inactive_page_gc = make_gc(pager->window, inactive_page_color);
 	pager->win_border_gc    = make_gc(pager->window, win_border_color);
+	pager->pager_border_gc  = make_gc(pager->window, pager_border_color);
 	pager->grid_gc          = make_gc(pager->window, grid_color);
+	pager->desk_grid_gc     = make_gc(pager->window, desk_grid_color);
 
 	visual = DefaultVisual(display, DefaultScreen(display));
 	cm = DefaultColormap(display, DefaultScreen(display));
@@ -958,8 +1067,8 @@ void pager_delete(struct pager *pager)
 
 	XFreeGC(display, pager->active_win_gc);
 	XFreeGC(display, pager->inactive_win_gc);
-	XFreeGC(display, pager->active_desk_gc);
-	XFreeGC(display, pager->inactive_desk_gc);
+	XFreeGC(display, pager->active_page_gc);
+	XFreeGC(display, pager->inactive_page_gc);
 	XFreeGC(display, pager->win_border_gc);
 	XFreeGC(display, pager->grid_gc);
 
@@ -1071,6 +1180,29 @@ void pager_set_allow_cover(struct pager *pager, int on)
 	pager->allow_cover = on != 0;
 }
 
+void pager_raise(struct pager *pager)
+{
+	x_window_set_below(pager->window, _NET_WM_STATE_REMOVE);
+	x_window_set_above(pager->window, _NET_WM_STATE_ADD);
+  	if (auto_hide) {
+	  	XMapWindow(display, pager->window);
+		XLowerWindow(display, pager->window);
+	}
+	XFlush(display);
+}
+
+void pager_lower(struct pager *pager)
+{
+	x_window_set_above(pager->window, _NET_WM_STATE_REMOVE);
+	x_window_set_below(pager->window, _NET_WM_STATE_ADD);
+	XFlush(display);
+}
+
+void pager_unmap(struct pager *pager)
+{
+	XUnmapWindow(display, pager->window);
+}
+
 void pager_show(struct pager *pager)
 {
 	pager_configure(pager);
@@ -1083,65 +1215,124 @@ void pager_show(struct pager *pager)
 
 	x_window_set_skip_pager(pager->window, _NET_WM_STATE_ADD);
 	x_window_set_skip_taskbar(pager->window, _NET_WM_STATE_ADD);
+	if (opt_tray)
+		x_request_systray(pager->window);
+
 	switch (pager->layer) {
-	case LAYER_BELOW:
-		x_window_set_above(pager->window, _NET_WM_STATE_REMOVE);
-		x_window_set_below(pager->window, _NET_WM_STATE_ADD);
-		break;
 	case LAYER_NORMAL:
 		break;
+	case LAYER_COMMUTE:
+	case LAYER_BELOW:
+		pager_lower(pager);
+		break;
 	case LAYER_ABOVE:
-		x_window_set_below(pager->window, _NET_WM_STATE_REMOVE);
-		x_window_set_above(pager->window, _NET_WM_STATE_ADD);
+		pager_raise(pager);
 		break;
 	}
-	x_window_set_desktop(pager->window, -1);
+	x_window_set_desktop(pager->window, -1, pager->size);
 }
 
 void pager_button_press(struct pager *pager, int x, int y, int button)
 {
-	int rx, ry, desk;
+	int rx, ry, active;
 
 	if (pager->popup_visible)
 		popup_hide(pager);
-	if (button != 1 && button != 2)
+	if (button < 1 || button > 5)
 		return;
+
 	if (pager->mouse.button != -1)
 		return;
 
+	pager_coords_to_root(pager, x, y, &rx, &ry, &active);
+        d_print("button press page=%d (%d %d) (%d %d)\n", active, x, y, rx, ry);
+
 	pager->mouse.button = button;
 	pager->mouse.dragging = 0;
 	pager->mouse.click_x = x;
 	pager->mouse.click_y = y;
 
-	pager_coords_to_root(pager, x, y, &rx, &ry, &desk);
-
-	pager->mouse.window_idx = get_window_idx(pager, rx, ry, desk);
+	pager->mouse.window_idx = get_window_idx(pager, rx, ry, active);
 	if (pager->mouse.window_idx != -1) {
 		pager->mouse.window_x = rx - pager->windows[pager->mouse.window_idx].x;
 		pager->mouse.window_y = ry - pager->windows[pager->mouse.window_idx].y;
 	}
-
 }
 
 void pager_button_release(struct pager *pager, int x, int y, int button)
 {
-	int rx, ry, desk;
+#if DEBUG > 0
+	int desk;
+#endif
+	int rx, ry, col, row, active, bool;
 
 	if (pager->mouse.button != button)
 		return;
 
-	pager_coords_to_root(pager, x, y, &rx, &ry, &desk);
+	pager_coords_to_root(pager, x, y, &rx, &ry, &active);
+	col = active % pager->size[COLS];
+	row = active / pager->size[COLS];
+	col = col / pager->size[PAGE_COLS];
+	row = row / pager->size[PAGE_ROWS];
+#if DEBUG > 0
+	desk = row * pager->size[DESK_COLS] + col;
+#endif
+        bool = (active != pager->active_page) && switch_delay;
 
-	if (button == 1) {
-		if (pager->mouse.window_idx == -1) {
-			x_set_current_desktop(desk);
-		} else if (!pager->mouse.dragging) {
-			x_set_active_window(pager->windows[pager->mouse.window_idx].window, SOURCE_INDICATION_PAGER);
+        d_print("button release page=%d desk=%d pager_xy (%d %d) root_xy (%d %d)\n", 
+		active, desk, x, y, rx, ry);
+
+	if (button == action_button[1]) {
+		if (pager->active_page != active) {
+			pager->active_page = active;
+			x_set_current_desktop(pager->active_page, pager->size);
 		}
-	} else if (button == 2) {
-	}
-
+		if (pager->mouse.window_idx != -1 && !pager->mouse.dragging) {
+			if (bool) {
+				usleep(500*switch_delay);
+				XSync(display, True);
+			}
+			x_set_current_desktop(pager->active_page, pager->size);
+                        x_set_active_window(pager->windows[pager->mouse.window_idx].window, SOURCE_INDICATION_PAGER);
+			XSetInputFocus(display, pager->windows[pager->mouse.window_idx].window, RevertToParent, CurrentTime);
+			if (bool) {
+				XSync(display, True);
+				usleep(500*switch_delay);
+				XSync(display, True);
+			}
+		}
+		pager->needs_update = 1;
+	} else if (button == action_button[2] || button >= 4) {
+	  	if (opt_transient && button == action_button[2]) {
+			exit(0);
+		}
+	  	if (opt_tray && button == action_button[2]) {
+			if (system("netwmpager -transient &"));
+			/* hack to silence warning: ignoring return value */
+			goto terminate;
+		} else
+	  	if (button == 4)
+			pager->zoom *= zoom_factor;
+		else
+		if (button == 5) {
+			pager->zoom /= zoom_factor;
+			if (pager->zoom < 0.5)
+				pager->zoom = 0.5;
+		} else
+	  	if (pager->zoom > 1.0)
+			pager->zoom = 1.0;
+		else
+			pager->zoom = zoom;
+		d_print("zoom: %g\n", pager->zoom);
+		pager_calc_w_h(pager);
+		XResizeWindow(display, pager->window, pager->w, pager->h);
+	terminate:
+		(void) set_pager_grid(pager, cols_init, rows_init, desk_cols_init, desk_rows_init);
+		pager->needs_configure = 1;
+	} else if (button == action_button[3]) {
+		if (pager->mouse.window_idx != -1)
+			pager->needs_update = 1;
+	} 
 	pager->mouse.window_idx = -1;
 	pager->mouse.button = -1;
 	pager->mouse.dragging = 0;
@@ -1149,9 +1340,15 @@ void pager_button_release(struct pager *pager, int x, int y, int button)
 
 void pager_motion(struct pager *pager, int x, int y)
 {
-	int rx, ry, desk;
+	int wx, wy, rx, ry, active_col, active_row, page;
+        static int init = 1;
 	struct client_window *window;
 
+	active_col = pager->active_page % pager->size[COLS];
+	active_row = pager->active_page / pager->size[COLS];
+	active_col = active_col % pager->size[PAGE_COLS];
+	active_row = active_row % pager->size[PAGE_COLS];
+
 	if (pager->mouse.button == -1) {
 		/* show / hide popup */
 		int idx;
@@ -1164,8 +1361,9 @@ void pager_motion(struct pager *pager, int x, int y)
 			fprintf(stderr, "XTranslateCoordinates failed\n");
 			return;
 		}
-		pager_coords_to_root(pager, x, y, &rx, &ry, &desk);
-		idx = get_window_idx(pager, rx, ry, desk);
+		pager_coords_to_root(pager, x, y, &rx, &ry, &page);
+		idx = get_window_idx(pager, rx, ry, page);
+
 		if (pager->popup_visible) {
 			if (idx == -1) {
 				popup_hide(pager);
@@ -1187,37 +1385,98 @@ void pager_motion(struct pager *pager, int x, int y)
 		/* move window */
 		pager->mouse.dragging = 1;
 
-		pager_coords_to_root(pager, x, y, &rx, &ry, &desk);
-		window = &pager->windows[pager->mouse.window_idx];
-		if (desk != window->desk && window->desk != -1) {
-			x_window_set_desktop(window->window, desk);
-			window->desk = desk;
-		}
-		if (pager->mouse.button == 1) {
-			/* move to other desk (already done :)) */
-		} else if (pager->mouse.button == 2) {
-			/* exact placement */
-			int wx, wy;
+		pager_coords_to_root(pager, x, y, &rx, &ry, &page);
 
+		window = &pager->windows[pager->mouse.window_idx];
+		if (pager->mouse.button == action_button[1]) {
+			/* move to other page */
 			/* can't set geometry of a shaded window */
 			if (window->states & WINDOW_STATE_SHADED)
 				x_window_set_shaded(window->window, _NET_WM_STATE_REMOVE);
+			repeat:
+			window->x += win_drift_x;
+			window->y += win_drift_y;
+			wx = window->x / pager->root_w;
+			rx = rx / pager->root_w - active_col;
+			wy = window->y / pager->root_h;
+			ry = ry / pager->root_h - active_row;
+                        while (window->x < (pager->size[PAGE_COLS]-active_col-1) * pager->root_w - window->w && rx > wx) {
+				window->x += pager->root_w;
+				++wx;
+			}
+                        while (window->x >= (1 - active_col) * pager->root_w && rx < wx) {
+				window->x -= pager->root_w;
+				--wx;
+			}
+                        while (window->y < (pager->size[PAGE_ROWS]-active_row-1) * pager->root_h - window->h && ry > wy) {
+				window->y += pager->root_h;
+				++wy;
+			}
+                        while (window->y >= (1 - active_row) * pager->root_h && ry < wy) {
+				window->y -= pager->root_h;
+				--wy;
+			}
+			if (page != window->page && window->page != -1) {
+				x_window_set_desktop(window->window, page, pager->size);
+				window->page = page;
+			}
+			x_window_set_geometry(window->window, XValue | YValue, window->x, window->y, 0, 0);
+			wx = window->x;
+			wy = window->y;
+			x_window_get_desktop(window->window, active_col, active_row, &window->page, &window->x, &window->y, &rx, &ry, pager->size);
+			/* accommodate strange behaviour of some WMs
+			   like compiz, fluxbox ... */
+			win_drift_x = wx - window->x;
+			win_drift_y = wy - window->y; 
+			if (init && (win_drift_x!=0 || win_drift_y !=0)) {
+				init = 0;
+				window->x += win_drift_x;
+				window->y += win_drift_y;
+				goto repeat;
+			}
 
-			wx = rx - pager->mouse.window_x;
-			wy = ry - pager->mouse.window_y;
+			d_print("moving window '%s' to (%d,%d) page: %d\n",
+				window->name, window->x, window->y,
+				window->page);
+			pager_update(pager);
+		} else if (pager->mouse.button == action_button[3]) {
+			/* exact placement */
+			/* can't set geometry of a shaded window */
+			if (window->states & WINDOW_STATE_SHADED)
+				x_window_set_shaded(window->window, _NET_WM_STATE_REMOVE);
+			wx = rx - pager->mouse.window_x - active_col * pager->root_w;
+			wy = ry - pager->mouse.window_y - active_row * pager->root_h;
 			x_window_set_geometry(window->window, XValue | YValue, wx, wy, 0, 0);
 			window->x = wx;
 			window->y = wy;
+			if (page != window->page && window->page != -1) {
+				x_window_set_desktop(window->window, page, pager->size);
+				window->page = page;
+			}
+			x_window_get_desktop(window->window, active_col, active_row, &window->page, &window->x, &window->y, &wx, &wy, pager->size);
+
+			d_print("moving window '%s' to (%d,%d) page: %d\n",
+				window->name, window->x, window->y,
+				window->page);
+			pager_update(pager);
 		}
 	}
 }
 
 void pager_enter(struct pager *pager, int x, int y)
 {
+	if (pager->layer == LAYER_COMMUTE) {
+		x_window_set_above(pager->window, _NET_WM_STATE_ADD);
+		x_window_set_below(pager->window, _NET_WM_STATE_REMOVE);
+	}
 	pager_motion(pager, x, y);
 }
 
 void pager_leave(struct pager *pager)
 {
-	popup_hide(pager);
+	/* popup_hide(pager); */
+	if (pager->layer == LAYER_COMMUTE) {
+		x_window_set_above(pager->window, _NET_WM_STATE_REMOVE);
+		x_window_set_below(pager->window, _NET_WM_STATE_ADD);
+	}
 }
