@@ -18,6 +18,7 @@
  */
 
 #include <x.h>
+#include <unistd.h>
 #include <xmalloc.h>
 #include <debug.h>
 
@@ -48,15 +49,18 @@ static const char *atom_names[NR_ATOMS] = {
 	"_NET_ACTIVE_WINDOW",
 	"_NET_CLIENT_LIST",
 	"_NET_CLIENT_LIST_STACKING",
+	"_NET_CLOSE_WINDOW",
 	"_NET_CURRENT_DESKTOP",
 	"_NET_DESKTOP_GEOMETRY",
 	"_NET_DESKTOP_LAYOUT",
 	"_NET_DESKTOP_NAMES",
 	"_NET_DESKTOP_VIEWPORT",
+	"_NET_MOVERESIZE_WINDOW",
 	"_NET_NUMBER_OF_DESKTOPS",
 	"_NET_SHOWING_DESKTOP",
 	"_NET_SUPPORTED",
 	"_NET_SUPPORTING_WM_CHECK",
+	"_NET_SYSTEM_TRAY_OPCODE",
 	"_NET_VIRTUAL_ROOTS",
 	"_NET_WM_ALLOWED_ACTIONS",
 	"_NET_WM_DESKTOP",
@@ -64,7 +68,6 @@ static const char *atom_names[NR_ATOMS] = {
 	"_NET_WM_ICON",
 	"_NET_WM_ICON_GEOMETRY",
 	"_NET_WM_ICON_NAME",
-	"_NET_CLOSE_WINDOW",
 	"_NET_WM_MOVERESIZE",
 	"_NET_WM_NAME",
 	"_NET_WM_PID",
@@ -104,6 +107,7 @@ static const char *atom_names[NR_ATOMS] = {
 };
 
 static Atom atom_values[NR_ATOMS] = { 0, };
+static Window x_root = None;
 
 Display *display = NULL;
 
@@ -241,6 +245,7 @@ int x_init(const char *display_name)
 	display = XOpenDisplay(display_name);
 	if (display == NULL)
 		return 1;
+        x_root = DefaultRootWindow(display); 
 	return 0;
 }
 
@@ -317,16 +322,16 @@ int x_is_netwm_compatible_wm_running(void)
 {
 	Window win;
 
-	if (x_get_window_property_nr(DefaultRootWindow(display), x_get_atom(_NET_SUPPORTING_WM_CHECK), &win, 1))
-		return 0;
-	return 1;
+	if (x_get_window_property_nr(x_root, x_get_atom(_NET_SUPPORTING_WM_CHECK), &win, 1))
+		return 1;
+	return 0;
 }
 
 int x_get_wm_name(char **wm_name)
 {
 	Window win;
 
-	if (x_get_window_property_nr(DefaultRootWindow(display), x_get_atom(_NET_SUPPORTING_WM_CHECK), &win, 1))
+	if (x_get_window_property_nr(x_root, x_get_atom(_NET_SUPPORTING_WM_CHECK), &win, 1))
 		return -1;
 	return x_get_utf8_string_property(win, x_get_atom(_NET_WM_NAME), wm_name);
 }
@@ -340,7 +345,7 @@ int x_get_client_list(int stacking, Window **windows, int *nr_windows)
 	} else {
 		property = x_get_atom(_NET_CLIENT_LIST);
 	}
-	return x_get_window_property(DefaultRootWindow(display), property, windows, nr_windows);
+	return x_get_window_property(x_root, property, windows, nr_windows);
 }
 
 int x_get_client_list_for_desktop(int stacking, int desktop, Window **windows, int *nr_windows)
@@ -355,9 +360,15 @@ int x_get_client_list_for_desktop(int stacking, int desktop, Window **windows, i
 	w = xnew(Window, nr);
 	j = 0;
 	for (i = 0; i < nr; i++) {
-		int desk;
+		unsigned long desk;
 
-		if (x_window_get_desktop(all_w[i], &desk) || desk != desktop)
+		if (x_get_cardinal_property_nr(all_w[i], x_get_atom(_NET_WM_DESKTOP), &desk, 1))
+			continue;
+
+		if (desk == 0xffffffff)
+			desk = -1;
+
+		if ((int)desk != desktop)
 			continue;
 		w[j++] = all_w[i];
 	}
@@ -369,14 +380,14 @@ int x_get_client_list_for_desktop(int stacking, int desktop, Window **windows, i
 	return 0;
 }
 
-int x_set_desktop_layout(int columns, int rows)
+int x_set_desktop_layout(int columns, int rows, int width, int height)
 {
 	unsigned long data[4];
 
 	if (columns < 1 || rows < 1)
 		return 1;
 
-	if (client_msg(DefaultRootWindow(display), DefaultRootWindow(display),
+	if (client_msg(x_root, x_root,
 			_NET_NUMBER_OF_DESKTOPS, columns * rows, 0, 0, 0, 0))
 		return -1;
 
@@ -384,25 +395,39 @@ int x_set_desktop_layout(int columns, int rows)
 	data[1] = columns;
 	data[2] = rows;
 	data[3] = _NET_WM_TOPLEFT;
-	if (x_set_cardinal_property(DefaultRootWindow(display), x_get_atom(_NET_DESKTOP_LAYOUT), data, 4))
+	if (x_set_cardinal_property(x_root, x_get_atom(_NET_DESKTOP_LAYOUT), data, 4))
 		return -1;
 	return 0;
 }
 
-int x_get_desktop_layout(int *columns, int *rows)
+int x_get_desktop_layout(int *columns, int *rows, int *width, int *height)
 {
 	unsigned long layout[4];
 	unsigned long nr_desks;
-	int c, r, properties_set_by_idiot = 0;
+	unsigned long *tmp;
+	int c, r, properties_set_by_idiot = 0, nr;
+
+	*width = DisplayWidth(display, DefaultScreen(display));
+	*height = DisplayHeight(display, DefaultScreen(display));
+	tmp = NULL;
+	(void) x_get_cardinal_property(x_root, x_get_atom(_NET_DESKTOP_GEOMETRY), &tmp, &nr);
+	if (nr >= 2 && tmp) {
+		*width = tmp[0];
+		*height = tmp[1];
+		if (*width <= 0)
+			*width = DisplayWidth(display, DefaultScreen(display));
+		if (*height <= 0)
+			*width = DisplayHeight(display, DefaultScreen(display));
+	}
 
 	/* this is set by WM */
-	if (x_get_cardinal_property_nr(DefaultRootWindow(display), x_get_atom(_NET_NUMBER_OF_DESKTOPS), &nr_desks, 1))
+	if (x_get_cardinal_property_nr(x_root, x_get_atom(_NET_NUMBER_OF_DESKTOPS), &nr_desks, 1))
 		return -1;
 	if (nr_desks == 0)
 		return -1;
 
 	/* this is set by pager */
-	if (x_get_cardinal_property_nr(DefaultRootWindow(display), x_get_atom(_NET_DESKTOP_LAYOUT), layout, 4)) {
+	if (x_get_cardinal_property_nr(x_root, x_get_atom(_NET_DESKTOP_LAYOUT), layout, 4)) {
 		c = nr_desks;
 		r = 1;
 		properties_set_by_idiot = 1;
@@ -436,7 +461,7 @@ int x_get_desktop_layout(int *columns, int *rows)
 		}
 	}
 	if (properties_set_by_idiot)
-		x_set_desktop_layout(c, r);
+		x_set_desktop_layout(c, r, *width, *height);
 
 	*columns = c;
 	*rows = r;
@@ -447,12 +472,12 @@ int x_get_desktop_names(char ***namesp)
 {
 	int nr;
 
-	return x_get_utf8_string_array_property(DefaultRootWindow(display), x_get_atom(_NET_DESKTOP_NAMES), namesp, &nr);
+	return x_get_utf8_string_array_property(x_root, x_get_atom(_NET_DESKTOP_NAMES), namesp, &nr);
 }
 
 int x_set_showing_desktop(int on)
 {
-	return client_msg(DefaultRootWindow(display), DefaultRootWindow(display),
+	return client_msg(x_root, x_root,
 			_NET_SHOWING_DESKTOP, on != 0, 0, 0, 0, 0);
 }
 
@@ -460,39 +485,79 @@ int x_get_showing_desktop(int *on)
 {
 	unsigned long state;
 
-	if (x_get_cardinal_property_nr(DefaultRootWindow(display), x_get_atom(_NET_SHOWING_DESKTOP), &state, 1))
+	if (x_get_cardinal_property_nr(x_root, x_get_atom(_NET_SHOWING_DESKTOP), &state, 1))
 		return -1;
 	*on = state != 0L;
 	return 0;
 }
 
-int x_set_current_desktop(int desktop)
+int x_set_current_desktop(int page, int *size)
 {
-	return client_msg(DefaultRootWindow(display), DefaultRootWindow(display),
-			_NET_CURRENT_DESKTOP, desktop, 0, 0, 0, 0);
+	int ret = 0, desk, col, row, desk_col, desk_row, w, h;
+	
+	row = page / size[COLS];
+	col = page - size[COLS] * row;
+
+        desk_col = col / size[PAGE_COLS];
+	col = col % size[PAGE_COLS];
+        desk_row = row / size[PAGE_ROWS];
+	row = row % size[PAGE_ROWS];
+
+	desk = desk_row * size[DESK_COLS] + desk_col;
+
+	ret = client_msg(x_root, x_root,
+			  _NET_CURRENT_DESKTOP, desk, 0, 0, 0, 0);
+	
+	w = DisplayWidth(display, DefaultScreen(display));
+	h = DisplayHeight(display, DefaultScreen(display));
+
+	d_print("set current page=%d (col,row)=(%d,%d) (desk_col,desk_row)=(%d,%d)\n", page, col, row, desk_col, desk_row);
+
+	ret |= client_msg(x_root, x_root,
+			  _NET_DESKTOP_VIEWPORT, col*w, row*h, 0, 0, 0);
+	return ret;
 }
 
-int x_get_current_desktop(int *desktop)
+int x_get_current_desktop(int *page, int *size)
 {
-	unsigned long tmp;
+	unsigned long desk=0, *tmp;
+	int nr, col=0, row=0, desk_col, desk_row, w, h;
 
-	if (x_get_cardinal_property_nr(DefaultRootWindow(display), x_get_atom(_NET_CURRENT_DESKTOP), &tmp, 1))
+	if (x_get_cardinal_property_nr(x_root, x_get_atom(_NET_CURRENT_DESKTOP), &desk, 1))
 		return -1;
-	*desktop = tmp;
+
+	if (x_get_cardinal_property(x_root, x_get_atom(_NET_DESKTOP_VIEWPORT), &tmp, &nr))
+		return -1;
+
+
+	if (nr >= 2) {
+		w = DisplayWidth(display, DefaultScreen(display));
+		h = DisplayHeight(display, DefaultScreen(display));
+		col = tmp[0]/w;
+		row = tmp[1]/h;
+	}
+
+	desk_col = desk % size[DESK_COLS];
+	desk_row = desk / size[DESK_COLS];
+
+	col = col + desk_col * size[PAGE_COLS];
+	row = row + desk_row * size[PAGE_ROWS];
+
+	*page = row * size[COLS] + col;
 	return 0;
 }
 
 int x_set_active_window(Window window, enum source_indication si)
 {
 	/* FIXME: [1] timestamp, [2] requestor's currently active window */
-	return client_msg(DefaultRootWindow(display), window, _NET_ACTIVE_WINDOW, si, 0, 0, 0, 0);
+	return client_msg(x_root, window, _NET_ACTIVE_WINDOW, si, 0, 0, 0, 0);
 }
 
 int x_get_active_window(Window *window)
 {
 	int rc;
 
-	rc = x_get_window_property_nr(DefaultRootWindow(display), x_get_atom(_NET_ACTIVE_WINDOW), window, 1);
+	rc = x_get_window_property_nr(x_root, x_get_atom(_NET_ACTIVE_WINDOW), window, 1);
 	if (rc == 0 && *window == 0)
 		return -2;
 	return rc;
@@ -500,33 +565,74 @@ int x_get_active_window(Window *window)
 
 int x_get_root_pixmap(Pixmap *pixmap)
 {
-	return x_get_property_nr(DefaultRootWindow(display), XA_PIXMAP, x_get_atom(_XROOTPMAP_ID), sizeof(Pixmap), pixmap, 1);
+	return x_get_property_nr(x_root, XA_PIXMAP, x_get_atom(_XROOTPMAP_ID), sizeof(Pixmap), pixmap, 1);
 }
 
-int x_window_set_desktop(Window window, int desktop)
+int x_window_set_desktop(Window window, int page, int *size)
 {
 	unsigned long desk;
+	int col, row, desk_col, desk_row;
 
-	if (desktop == -1) {
+	if (page == -1) {
 		desk = 0xffffffff;
-	} else {
-		desk = desktop;
 	}
-	return client_msg(DefaultRootWindow(display), window, _NET_WM_DESKTOP, desk, 0, 0, 0, 0);
+
+	col = page % size[COLS];
+	row = page / size[COLS];
+
+	desk_col = col / size[PAGE_COLS];
+	col = col % size[PAGE_COLS];
+	desk_row = row / size[PAGE_ROWS];
+	row = row % size[PAGE_ROWS];
+	desk = desk_row * size[DESK_COLS] + desk_col;
+
+	return client_msg(x_root, window, _NET_WM_DESKTOP, desk, 0, 0, 0, 0);
 }
 
-int x_window_get_desktop(Window window, int *desktop)
+int x_window_get_desktop(Window window, int active_col, int active_row, int *page, int *px, int *py, int *pw, int *ph, int *size)
 {
-	unsigned long desk;
+	long unsigned int desk = 0;
+	int w, h, col, row, desk_col, desk_row;
 
 	if (x_get_cardinal_property_nr(window, x_get_atom(_NET_WM_DESKTOP), &desk, 1))
 		return -1;
 
-	if (desk == 0xffffffff) {
-		*desktop = -1;
-	} else {
-		*desktop = desk;
+	if (desk == 0xffffffff)
+		*page = -1;
+	else
+		*page = (int) desk;
+
+	if (x_window_get_geometry(window, px, py, pw, ph)) {
+		/* set simple default values and warn user */
+		*px = 0;
+		*py = 0;
+		*pw = 100;
+		*ph = 60;
+		fprintf(stderr, "could not get geometry of window 0x%x\n", (int)window);
 	}
+
+  	/* correct "raw" x, y to take account of active page 
+	   in case one has a virtual root containing
+	   several pages */
+	w = DisplayWidth(display, DefaultScreen(display));
+	h = DisplayHeight(display, DefaultScreen(display));
+	*px += active_col * w;
+	*py += active_row * h;
+
+	if (*page == -1)
+		return 0;
+
+	desk_col = ((int)desk) % size[DESK_COLS];
+	desk_row = ((int)desk) / size[DESK_COLS];
+
+	col = *px / w;
+	row = *py / h;
+
+	col += desk_col * size[PAGE_COLS];
+	row += desk_row * size[PAGE_ROWS];
+
+	*page = row * size[COLS] + col;
+
 	return 0;
 }
 
@@ -677,7 +783,9 @@ int x_window_get_geometry(Window window, int *x, int *y, int *w, int *h)
  * 		return -1;
  * 	}
  */
-	if (XTranslateCoordinates(display, window, DefaultRootWindow(display), a.x, a.y, x, y, &child_ret) == True) {
+	if (XTranslateCoordinates(display, x_root, window, a.x, a.y, x, y, &child_ret) == True) {
+	        *x = a.x - *x;
+	        *y = a.y - *y;
 		*w = a.width;
 		*h = a.height;
 	} else {
@@ -791,7 +899,7 @@ int x_window_set_states(Window window, enum state_action action, enum atom_index
 	} else {
 		a2 = x_get_atom(prop2);
 	}
-	return client_msg(DefaultRootWindow(display), window, _NET_WM_STATE, action, a1, a2, 0, 0);
+	return client_msg(x_root, window, _NET_WM_STATE, action, a1, a2, 0, 0);
 }
 
 int x_window_get_states(Window window, unsigned int *states)
@@ -858,15 +966,23 @@ int x_window_set_aspect(Window window, int x, int y)
 
 int x_window_close(Window window)
 {
-	return client_msg(DefaultRootWindow(display), window, _NET_CLOSE_WINDOW,
+	return client_msg(x_root, window, _NET_CLOSE_WINDOW,
 			0, 0, 0, 0, 0);
 }
 
 int x_window_move_resize(Window window,
+		unsigned long x, unsigned long y,
+		unsigned long width, unsigned long height)
+{
+	return client_msg(x_root, window, _NET_MOVERESIZE_WINDOW,
+			  0, x, y, width, height);
+}
+
+int x_window_initiate_move_resize(Window window,
 		unsigned long x_root_button, unsigned long y_root_button,
 		unsigned long direction, unsigned long button)
 {
-	return client_msg(DefaultRootWindow(display), window, _NET_WM_MOVERESIZE,
+	return client_msg(x_root, window, _NET_WM_MOVERESIZE,
 			x_root_button, y_root_button, direction, button, 0);
 }
 
@@ -981,3 +1097,28 @@ int x_window_set_undecorated(Window window, enum state_action action)
 {
 	return x_window_set_states(window, action, _OB_WM_STATE_UNDECORATED, 0);
 }
+
+void x_request_systray(Window window)
+{
+	XEvent event;
+	Window tray_window;
+	char systray[32];
+
+	sprintf(systray, "_NET_SYSTEM_TRAY_S%d", DefaultScreen(display));
+	tray_window = XGetSelectionOwner(display, XInternAtom (display, systray, False ));	
+
+	memset(&event, 0, sizeof(event));
+	event.xclient.type = ClientMessage;
+	event.xclient.window = tray_window;
+	event.xclient.message_type = x_get_atom(_NET_SYSTEM_TRAY_OPCODE);
+	event.xclient.format = 32;
+
+	event.xclient.data.l[0] = CurrentTime;
+	event.xclient.data.l[1] = 0;
+	event.xclient.data.l[2] = window;
+	event.xclient.data.l[3] = 0;
+	event.xclient.data.l[4] = 0;
+
+	XSendEvent(display, tray_window, False, NoEventMask, &event);
+        XSync(display, False);
+}
